import sys
if sys.version_info < (3,5):
    raise EnvironmentError('Please install a python version >= 3.6 to use this library')

from mindsdb_sdk.classes.sdk import SDK
from mindsdb_sdk.classes.pandas_acessor import AutoML, auto_ml_config



from mindsdb_sdk.helpers.net_helpers import sending_attempts
from mindsdb_sdk.helpers.exceptions import IntegrationException


class Integration:
    def __init__(self, proxy, name):
        self._proxy = proxy
        self.name = name
        self._route = "/config/integrations"

    @sending_attempts(exception_type=IntegrationException)
    def get_info(self):
        return self._proxy.get(f'{self._route}/{self.name}')

    def update(self, params):
        """Update existing integration config.
        params -> dict: new set of integration parameters.
        {"params":{
            "enabled": true,
            "host": "localhost",
            "password": "mypass",
            "port": 3306,
            "type": "mariadb",
            "user": "root"
        }
        }
        """
        return self._proxy.post(f'{self._route}/{self.name}', json=params)

    def __len__(self):
        return len(self.get_info())


class Integrations:
    def __init__(self, proxy):
        self._proxy = proxy
        self._route = '/config/integrations'

    @sending_attempts(exception_type=IntegrationException)
    def list_info(self):
        return self._proxy.get(self._route)

    def list_integrations(self):
        return self.list_info()["integrations"]

    def __len__(self):
        return len(self.list_integrations())

    def __getitem__(self, name):
        if name in self.list_integrations():
            return Integration(self._proxy, name)
        return None

    def __setitem__(self, name, params):
        """Creates new integration config.
        params -> dict: new set of integration parameters.
        {"params":{
            "enabled": true,
            "host": "localhost",
            "password": "mypass",
            "port": 3306,
            "type": "mariadb",
            "user": "root"
        }
        }
        """
        self._proxy.put(f'{self._route}/{name}', json=params)

    def __delitem__(self, name):
        self._proxy.delete(f'{self._route}/{name}')





import os
import pandas as pd

from mindsdb_sdk.classes import proxy
from mindsdb_sdk.classes import predictors
from mindsdb_sdk.classes import datasources


def auto_ml_config(mode='native', connection_info=None):
    if mode == 'native':
        os.environ['MINDSDB_PANDAS_AUTOML_MODE'] = 'native'
    elif mode == 'api':
        os.environ['MINDSDB_PANDAS_AUTOML_MODE'] = 'api'
        os.environ['MINDSDB_PANDAS_AUTOML_HOST'] = connection_info['host']
        # not gracefully but as temporary solution
        # need to reset auth env for each instance with 'api' case
        # to prevent collision
        for env in ('MINDSDB_PANDAS_AUTOML_USER', 'MINDSDB_PANDAS_AUTOML_PASSWORD', 'MINDSDB_PANDAS_AUTOML_TOKEN'):
            try:
                del os.environ[env]
            except KeyError:
                pass

        if 'user' in connection_info:
            os.environ['MINDSDB_PANDAS_AUTOML_USER'] = connection_info['user']

        if 'password' in connection_info:
            os.environ['MINDSDB_PANDAS_AUTOML_PASSWORD'] = connection_info['password']

        if 'token' in connection_info:
            os.environ['MINDSDB_PANDAS_AUTOML_TOKEN'] = connection_info['token']
    else:
        raise Exception(f'Invalid mode: {mode} for the pandas auto_ml accessor!')


@pd.api.extensions.register_dataframe_accessor("auto_ml")
class AutoML:
    def __init__(self, pandas_obj):
        self._df = pandas_obj
        self._predictor = None
        self._analysis = None
        self.mode = os.environ['MINDSDB_PANDAS_AUTOML_MODE']
        self._raw_name = str(pd.util.hash_pandas_object(self._df).sum())

        if self.mode == 'api':
            self.host = os.environ['MINDSDB_PANDAS_AUTOML_HOST']
            self.user = os.environ.get('MINDSDB_PANDAS_AUTOML_USER', None)
            self.password = os.environ.get('MINDSDB_PANDAS_AUTOML_PASSWORD', None)
            self.token = os.environ.get('MINDSDB_PANDAS_AUTOML_TOKEN', None)
            self.proxy = proxy.Proxy(self.host, user=self.user, password=self.password, token=self.token)
            self.remote_datasource_controller = datasources.DataSources(self.proxy)
            self.predictor_class = predictors.Predictors(self.proxy)
        else:
            from mindsdb_native import Predictor as NativePredictor
            self.predictor_class = NativePredictor

    @property
    def analysis(self):
        if self._analysis is not None:
            return self._analysis
        if self.mode == 'native':
            from mindsdb_native.libs.controllers.functional import analyse_dataset
            self._analysis = analyse_dataset(self._df)
        else:
            datasource = self.remote_datasource_controller[self._raw_name]
            if datasource is None:
                self.remote_datasource_controller[self._raw_name] = {'df': self._df}
                datasource = self.remote_datasource_controller[self._raw_name]
            self._analysis = datasource.analyze()

        return self._analysis

    def learn(self, to_predict, name=None):

        if name is None:
            name = self._raw_name
        self._predictor = self.predictor_class(name)
        self._predictor.learn(from_data=self._df, to_predict=to_predict)

        return name


    def predict(self, name=None, when_data=None):

        if name is not None:
            predictor = self.predictor_class(name)
        else:
            predictor = self._predictor
        if when_data is None:
            return predictor.predict(when_data=self._df)
        return predictor.predict(when_data=when_data)



import os
import requests

from mindsdb_sdk.classes import authorizers


def for_status_raiser(func):
    """Decorator which raises exception if response code returned by 'func'
    is not 200"""
    def wrapper(*args, **kwargs):
        r = func(*args, **kwargs)
        try:
            r.raise_for_status()
        except Exception as e:
            raise Exception(f"Error with message: {r.text}") from e
        return r.json()
    return wrapper


class Proxy:

    def __init__(self, host, user=None, password=None, token=None, url_token=None):
        self._host = host.rstrip('/')
        self._apikey = None

        if url_token is not None:
            self._authorizer = authorizers.UrlTokenAuthorizer(host, user, password, url_token)
        elif (user is not None and password is not None) or token is not None:
            self._authorizer = authorizers.CloudAuthorizer(host, user, password, token=token)
        else:
            self._authorizer = authorizers.BaseAuthorizer(host, user, password, token=token)

    @for_status_raiser
    def post(self, route, data=None, json=None, params=None):
        if params is None:
            params = {}

        return self._authorizer('post',
                                self._host + '/api' + route,
                                data=data,
                                params=params,
                                json=json)

    @for_status_raiser
    def put(self, route, data=None, json=None, params=None, files=None, params_processing=True):

        if params_processing:
            if params is None:
                params = {}

            if files is not None:
                del_tmp = False
                if 'df' in files:
                    del_tmp = True
                    files['df'].to_csv('tmp_upload_file.csv', index=False)
                    files['file'] = 'tmp_upload_file.csv'
                    del files['df']

                with open(files['file'], 'rb') as fp:
                    files['file'] = fp
                    data = {}
                    data['source_type'] = 'file'

                    response = self._authorizer('put',
                                                self._host + '/api' + route,
                                                files=files,
                                                data=data)

                if del_tmp:
                    os.remove('tmp_upload_file.csv')

                return response


        response = self._authorizer('put',
                                    self._host + '/api' + route,
                                    data=data,
                                    params=params,
                                    json=json,
                                    files=files)


        return response

    @for_status_raiser
    def get(self, route, params=None):
        if params is None:
            params = {}

        return self._authorizer('get',
                                self._host + '/api' + route,
                                params=params)


    @for_status_raiser
    def delete(self, route, params=None):
        if params is None:
            params = {}

        return self._authorizer('delete',
                                self._host + '/api' + route,
                                params=params)


    def ping(self):
        try:
            return self.get('/util/ping')['status'] == 'ok'
        except Exception as _:
            return False



import requests


class BaseAuthorizer:
    def __init__(self, host, username, password, *args, **kwargs):
        self.host = host
        self.username = username
        self.password = password
        self.token = kwargs.get('token', None)

    def __call__(self, req_type, url,  **kwargs):
        return getattr(requests, req_type)(url, **kwargs)


class CloudAuthorizer(BaseAuthorizer):
    def __init__(self, host, username, password, *args, **kwargs):
        super().__init__(host, username, password, *args, **kwargs)
        self.base_url = f"{self.host}/cloud"
        self.token = self._get_api_token()

    def _get_api_token(self):
        if self.token is not None:
            return self.token
        token_url = self.base_url + '/login'
        json = {'email': self.username, 'password': self.password}
        r = requests.post(token_url, json=json)
        r.raise_for_status()
        return r.cookies['apiKey']

    @property
    def auth_cookies(self):
        return {'apiKey': self.token}


    def __call__(self, req_type, url,  **kwargs):
        kwargs['cookies'] = self.auth_cookies
        return getattr(requests, req_type)(url, **kwargs)


class UrlTokenAuthorizer(BaseAuthorizer):
    def __init__(self, host, username, password, url_token):
        super().__init__(host, username, password)
        self.token = url_token

    def __call__(self, req_type, url,  **kwargs):
        params = kwargs.get('params', {})
        params['apikey'] = self.token
        kwargs['params'] = params
        return getattr(requests, req_type)(url, **kwargs)



import time
import os
from tempfile import NamedTemporaryFile
from mindsdb_sdk.helpers.net_helpers import sending_attempts
from mindsdb_sdk.helpers.exceptions import DataSourceException


class DataSource():
    def __init__(self, proxy, name):
        self._proxy = proxy
        self.name = name
        self._analysis = None

    @sending_attempts(exception_type=DataSourceException)
    def get_info(self):
        return self._proxy.get(f'/datasources/{self.name}')

    @sending_attempts(exception_type=DataSourceException, delay=1)
    def _get_analyze_data(self):
        return self._proxy.get(f'/datasources/{self.name}/analyze')

    def analyze(self, wait_seconds=360):
        if self._analysis is None:
            threshold = time.time() + wait_seconds
            analysis = self._get_analyze_data()
            while time.time() < threshold:
                if 'status' in analysis and analysis['status'] == 'analyzing':
                    time.sleep(10)
                analysis = self._get_analyze_data()

            if 'status' in analysis and analysis['status'] == 'analyzing':
                raise Exception(f'Analysis not yet ready after waiting for {wait_seconds}, consider setting the `wait_seconds` to a higher value or reporting a bug if you think this should not take as long as it does for your dataset.')
            self._analysis = analysis
        return self._analysis

    def __iter__(self):
        return iter(list(self.analyze().keys()))

    def __len__(self):
        return len(list(self.analyze().keys()))

    def __getitem__(self, k):
        return self.analyze()[k]

    def __delete__(self):
        self._proxy.delete(f'/datasources/{self.name}')


class DataSources():
    def __init__(self, proxy):
        self._proxy = proxy

    @sending_attempts(exception_type=DataSourceException)
    def list_info(self):
        return self._proxy.get('/datasources')

    def list_datasources(self):
        return [DataSource(self._proxy, x['name']) for x in self.list_info()]

    def __getitem__(self, name):
        datasources = (x['name'] for x in self.list_info())
        return DataSource(self._proxy, name) if name in datasources else None

    def __len__(self) -> int:
        return len(self.list_datasources())

    def __delitem__(self, name):
        self._proxy.delete(f'/datasources/{name}')

    def __setitem__(self, name, params):
        '''
        params is a dictionary that can contain:
        * file - File path
        * df - pandas dataframe
        * ulr - Url to file
        * source - file | url | <integration id>
        * source_type - file | ??

        and if source == <integration id>:
            * query
            + additional integration specific params (see docs in mindsdb)

        UPDATE: the general or (single) datasource type is pandas.DataFrame
        so only this type being handled in initial version
        '''
        files = {}
        data = {}
        for k in params:
            if k not in ['file', 'df', 'source', 'source_type']:
                data[k] = params[k]
            else:
                files[k] = params[k]

        if not files:
            files = None
            self._proxy.put(f'/datasources/{name}', files=files, data=data)
            return

        src_fd = None
        if 'df' in files:
            src_fd = NamedTemporaryFile(mode='w+', newline='')
            files['df'].to_csv(path_or_buf=src_fd, index=False)
            src_fd.flush()
            src_fd.seek(os.SEEK_SET)
            files['file'] = src_fd
            del files['df']
        if 'file' in files:
            src = files['file']
            if isinstance(src, str):
                src_fd = open(src, 'r')
            # check if it is a file-like object
            elif hasattr(src, 'read'):
                src_fd = src
            else:
                raise Exception(f"unknown type files['file']: {files['file']}")

        with src_fd:
            files['file'] = (src_fd.name.split('/')[-1], src_fd, 'text/csv')

            files['source_type'] = (None, 'file')
            files['source'] = (None, src_fd.name.split('/')[-1])

            files['name'] = (None, name)
            self._proxy.put(f'/datasources/{name}', files=files, data=data, params_processing=False)



"""This module contains set of solution-specific exceptions"""


class SDKException(Exception):
    def __init__(self, ret):
        self.ret = ret


class DataSourceException(SDKException):
    pass


class IntegrationException(SDKException):
    pass


class PredictorException(SDKException):
    pass


class AccessorException(SDKException):
    pass



__title__ = 'pd_auto_ml'
__package_name__ = 'pd_auto_ml'
__version__ = '0.4.0'
__description__ = "Pd auto_ml provides a pandas dataframe accessor to allow using dataframes with MindsDB (either the local native library or a local or remote mindsdb server)"
__email__ = "jorge@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/mindsdb_python_sdk'
__pypi__ = 'https://pypi.org/project/pd_auto_ml/'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2020- mindsdb'



import time
import pandas as pd
from pandas.util import hash_pandas_object
from mindsdb_sdk.classes.datasources import DataSources, DataSource
from mindsdb_sdk.helpers.net_helpers import sending_attempts
from mindsdb_sdk.helpers.exceptions import PredictorException, DataSourceException


class Predictor():
    def __init__(self, proxy, name):
        self._proxy = proxy
        self.name = name
        self.known_datasources = {}

    @staticmethod
    def _data_hash(data):
        if not isinstance(data, pd.DataFrame):
            raise Exception(f"invalid data type: {type(data)}. pandas.DataFrame expected")

        return hash_pandas_object(data).sum()

    @sending_attempts(exception_type=PredictorException)
    def get_info(self):
        return self._proxy.get(f'/predictors/{self.name}')

    def delete(self):
        self._proxy.delete(f'/predictors/{self.name}')

    def wait_readiness(self):
        while self.get_info()['status'] != 'complete':
            time.sleep(3)

    def predict(self, when_data, args=None):
        self.get_info()
        if isinstance(when_data, dict):
            json = when_data
            url = f'/predictors/{self.name}/predict'

        else:
            ds = self._check_datasource(when_data)
            json = {'data_source_name': ds.name}
            url = f'/predictors/{self.name}/predict_datasource'

        self.wait_readiness()
        return self._proxy.post(url, json=json)

    def _check_datasource(self, df):
        df_hash = self._data_hash(df)
        name = f"datasource_{df_hash}"
        datasource = DataSource(self._proxy, name)
        try:
            datasource.get_info()
        except DataSourceException:
            datasources = DataSources(self._proxy)
            datasources[name] = {'df': df}
        return datasource

    def learn(self, to_predict, from_data, args=None):
        if args is None:
            args = {}
        ds = self._check_datasource(from_data)

        self._proxy.put(f'/predictors/{self.name}', json={
            'data_source_name': ds.name,
            'kwargs': args,
            'to_predict': to_predict
        })


class Predictors():
    def __init__(self, proxy):
        self._proxy = proxy

    @sending_attempts(exception_type=PredictorException)
    def list_info(self):
        return self._proxy.get('/predictors')

    def  list_predictor(self):
        return [Predictor(self._proxy, x['name']) for x in self.list_info()]

    def __getitem__(self, name):
        predictors = (x['name'] for x in self.list_info())
        return Predictor(self._proxy, name) if name in predictors else None

    def __len__(self) -> int:
        return len(self.list_predictor())

    def __delitem__(self, name):
        self._proxy.delete(f'/predictors/{name}')

    def learn(self, name, datasource, to_predict, args=None):
        """Not sure that it is needed here. But left it now."""
        if args is None:
            args = {}
        datasource = datasource['name'] if isinstance(datasource, dict) else datasource
        self._proxy.put(f'/predictors/{name}', json={
            'data_source_name': datasource,
            'kwargs': args,
            'to_predict': to_predict
        })

    def __call__(self, name, **kwargs):
        return Predictor(self._proxy, name)


    '''
    @TODO:
    * Add custom predictor
    * Fit custom predictor
    * Upload predictor
    * Download predictor
    * Rename predictor
    '''
